#!/usr/bin/env python
# coding: utf-8

from kazoo.client import KazooClient

from k3confloader import conf
import k3utfjson

from . import zkutil


class ZKConf(object):
    """
    It is a config wrapper, provding several method for accessing config.
    If one of the config field is not spedified when initializing this class, it
    falls back to using `config.zk_<field>`.

    Classes in this module relies on this class to access config.

    E.g.:
    """

    # config.zk_journal_dir = "my_dir/"
    # c = k3zkutil.ZKConf(hosts="127.0.0.1:9999")
    # c.hosts()       # "127.0.0.1:9999" # specified
    # c.journal_dir() # "my_dir/"        # by default using `config.zk_<field>`
    """
    zk dir:
        <prefix>/record/<key>
        <prefix>/lock/<key>
        <prefix>/tx/
                    alive/0000000001
                    journal/0000000001
                    state/0000000001
                    journal_id_set
                    txid_maker
        <prefix>/seq/<key>

    alive:      Contains `ephemeral` node each of which represents a transaction.
                Its modifications is used.

    journal:    Contains journal transaction modifications. Each of them is a complete transaction.

    journal_id_set: Committed and Purged journal id.
    """

    def __init__(
        self, hosts=None, tx_dir=None, seq_dir=None, record_dir=None, lock_dir=None, node_id=None, auth=None, acl=None
    ):
        self.conf = {
            "hosts": hosts,
            "tx_dir": tx_dir,
            "seq_dir": seq_dir,
            "record_dir": record_dir,
            "lock_dir": lock_dir,
            "node_id": node_id,
            "auth": auth,
            "acl": acl,
        }

    def hosts(self):
        return self._get_config("hosts")

    def record_dir(self):
        return self._get_config("record_dir")

    def lock_dir(self):
        return self._get_config("lock_dir")

    def node_id(self):
        return self._get_config("node_id")

    def auth(self):
        return self._get_config("auth")

    def acl(self):
        return self._get_config("acl")

    def lock(self, key=""):
        return "".join([self.lock_dir(), _dump_txid(key)])

    def record(self, key=""):
        return "".join([self.record_dir(), key])

    def seq_dir(self):
        return self._get_config("seq_dir")

    def seq(self, key=""):
        return "".join([self.seq_dir(), key])

    def tx_dir(self):
        return self._get_config("tx_dir")

    def tx_alive(self, txid=""):
        return "".join([self.tx_dir(), "alive/", _dump_txid(txid)])

    def tx_state(self, txid=""):
        return "".join([self.tx_dir(), "state/", _dump_txid(txid)])

    def journal(self, journal_id=""):
        return "".join([self.tx_dir(), "journal/", _dump_journal_id(journal_id)])

    def journal_id_set(self):
        return "".join([self.tx_dir(), "journal_id_set"])

    def txid_maker(self):
        return "".join([self.tx_dir(), "txid_maker"])

    def kazoo_digest_acl(self):
        a = self.acl()
        if a is None:
            return a

        return zkutil.make_kazoo_digest_acl(a)

    def kazoo_auth(self):
        a = self.auth()
        if a is None:
            return None

        return a[0], a[1] + ":" + a[2]

    def _get_config(self, name):
        if self.conf[name] is None:
            return getattr(conf, "zk_" + name)
        else:
            return self.conf[name]


class KazooClientExt(KazooClient):
    def __init__(self, zkclient, json=True):
        if isinstance(zkclient, KazooClientExt):
            self._zk = zkclient._zk
            self._zkconf = ZKConf(**zkclient._zkconf.conf)

        elif isinstance(zkclient, KazooClient):
            self._zk = zkclient
            self._zkconf = None
        else:
            raise TypeError("invalid zkclient type: expect KazooClient or KazooClientExt")

        self._json = json

    def __getattr__(self, n):
        return getattr(self._zk, n)

    def _jl(self, v):
        if self._json:
            return k3utfjson.load(v)
        else:
            return v

    def _jd(self, v):
        if self._json:
            return k3utfjson.dump(v)
        else:
            return v

    def _encode(self, v):
        if isinstance(v, str):
            return v.encode("utf-8")
        return v

    def get(self, path, watch=None):
        val, zstat = self._zk.get(path, watch=watch)
        return self._jl(val), zstat

    def set(self, path, value, version=-1):
        value = self._jd(value)
        value = self._encode(value)
        return self._zk.set(path, value, version=version)

    def create(self, path, value=b"", acl=None, ephemeral=False, sequence=False, makepath=False):
        value = self._jd(value)

        return self._zk.create(path, value=value, acl=acl, ephemeral=ephemeral, sequence=sequence, makepath=makepath)


def _dump_txid(txid):
    if isinstance(txid, str):
        return txid
    elif isinstance(txid, int):
        return "%010d" % txid
    else:
        raise TypeError("invalid type txid: " + repr(txid))


def _dump_journal_id(journal_id):
    if isinstance(journal_id, int):
        return "journal_id%010d" % journal_id
    elif isinstance(journal_id, str):
        return journal_id
    else:
        raise TypeError("invalid type journal id: " + repr(journal_id))


def kazoo_client_ext(zk, json=True):
    """
    return zkclient created or original zkclient, and if zkclient is created
    """

    zkconf = None

    if isinstance(zk, str):
        zkconf = ZKConf(hosts=zk)

    if isinstance(zk, dict):
        zkconf = ZKConf(**zk)

    if isinstance(zk, ZKConf):
        zkconf = zk

    if zkconf is None:
        zkconf = ZKConf()
        owning = False
    else:
        zk = KazooClient(zkconf.hosts())
        owning = True

    zkclient = KazooClientExt(zk, json=json)

    if zkclient._zkconf is None:
        zkclient._zkconf = zkconf

    zkclient.start()

    auth = zkconf.kazoo_auth()
    if auth is not None:
        zkclient.add_auth(*auth)

    return zkclient, owning



"""
Some helper function to make life easier with zookeeper.

"""

from importlib.metadata import version

__version__ = version("k3zkutil")

from .exceptions import (
    ZKWaitTimeout,
)

from .zkacid import (
    cas_loop,
)

from .zkconf import (
    KazooClientExt,
    ZKConf,
    kazoo_client_ext,
)

from .zkutil import (
    PermTypeError,
    ZkPathError,
    close_zk,
    init_hierarchy,
    export_hierarchy,
    is_backward_locking,
    lock_id,
    make_acl_entry,
    make_digest,
    make_kazoo_digest_acl,
    parse_kazoo_acl,
    parse_lock_id,
    perm_to_long,
    perm_to_short,
    wait_absent,
    get_next,
)

from .zklock import (
    ZKLock,
    LockTimeout,
    make_identifier,
)

from .cached_reader import (
    CachedReader,
)

__all__ = [
    "PermTypeError",
    "ZKWaitTimeout",
    "ZkPathError",
    "cas_loop",
    "KazooClientExt",
    "ZKConf",
    "kazoo_client_ext",
    "close_zk",
    "init_hierarchy",
    "export_hierarchy",
    "is_backward_locking",
    "lock_id",
    "make_acl_entry",
    "make_digest",
    "make_kazoo_digest_acl",
    "parse_kazoo_acl",
    "parse_lock_id",
    "perm_to_long",
    "perm_to_short",
    "wait_absent",
    "get_next",
    "ZKLock",
    "LockTimeout",
    "CachedReader",
    "make_identifier",
]



import base64
import hashlib
import logging
import os
import threading
import time
import uuid
from collections import namedtuple

from kazoo import security
from kazoo.client import KazooClient
from kazoo.exceptions import NoNodeError
from kazoo.exceptions import KazooException
from k3confloader import conf
import k3net
import k3utfjson

from .exceptions import ZKWaitTimeout

logger = logging.getLogger(__name__)

PERM_TO_LONG = {
    "c": "create",
    "d": "delete",
    "r": "read",
    "w": "write",
    "a": "admin",
}

PERM_TO_SHORT = {
    "create": "c",
    "delete": "d",
    "read": "r",
    "write": "w",
    "admin": "a",
}

# We assumes that ip does not change during process running.
# Display intra ip if presents, or display pub ip.
host_ip4 = k3net.ips_prefer(k3net.get_host_ip4(), k3net.INN)


class PermTypeError(Exception):
    pass


class ZkPathError(Exception):
    pass


def close_zk(zk):
    """
    Stop and close a zk client.
    :param zk: a `KazooClient` or `KazooClientExt` instance, otherwise raise a `TypeError`.
    :return: nothing
    """
    if not isinstance(zk, KazooClient):
        raise TypeError("expect KazooClient or KazooClientExt, but got {t}".format(t=type(zk)))

    try:
        zk.stop()

    except KazooException as e:
        logger.exception(repr(e) + " while stop zk client")

    try:
        zk.close()

    except Exception as e:
        logger.exception(repr(e) + " while close zk client")


def lock_data(node_id=None):
    # deprecated
    return lock_id(node_id=node_id)


def lock_id(node_id=None):
    """
    It builds a string used as zk lock node value, to describe lock holder's information.
    The result string is in form:
    <node_id>-<ip>-<process_id>-<counter>
    # e.g. web-192.168.0.2-1233-0000000001

    `ip` is chosen from all local ipv4.
    If there is an intra net ip, use it.
    Otherwise, use an public ip it found.
    :param node_id: is an arbitrary string representing a host.
    If it is `None`, `config.zk_node_id` is used.

    :return: a string for lock identifier.
    """
    """
    Embed lock holder information into the zk node data for the lock.

    `node_id` is a user defined identifier of a host.
    """

    if node_id is None:
        node_id = conf.zk_node_id

    ip = (host_ip4 + ["unknownip"])[0]

    seq = [node_id, ip, str(os.getpid()), str(uuid.uuid4()).replace("-", "")]

    return "-".join(seq)


def parse_lock_data(data_str):
    # deprecated
    return parse_lock_id(data_str)


def parse_lock_id(data_str):
    """
    It parse string built by `zkutil.lock_id()` to an dictionary.
    :param data_str: string built by `zkutil.lock_id()`.
    """
    # :return:
    # {
    #     "node_id": "<node_id>",
    #     "ip": "<ip>",
    #     "process_id": process_id,
    #     "counter": <integer>,
    # "txid": "<txid>",
    # }
    """
    Parse string generated by lock_id()
    `process_id` and `counter` is `int`, the others are `string`s.

    If any of these three field can not be parsed correctly, it will be `None`.
    If `node_id` is in form of `txid:...`, `txid` is filled with the text after `:`.
    Otherwise it is `None`.
    This is added for zk-transaction implementation.
    """

    node_id, ip, process_id, _uuid = (data_str.split("-", 3) + ([None] * 4))[:4]

    if isinstance(process_id, str) and process_id.isdigit():
        process_id = int(process_id)
    else:
        process_id = None

    rst = {"node_id": node_id, "ip": ip, "process_id": process_id, "uuid": _uuid, "txid": None}

    if node_id.startswith("txid:"):
        rst["txid"] = node_id.split(":", 1)[1]

    return rst


def make_digest(acc):
    # acc = "username:password"
    """
    It make a digest for string acc

    :param acc: The `acc` string is in form:
    :return: a digest string.
    """
    digest = hashlib.sha1(acc.encode()).digest()
    rst = base64.b64encode(digest).strip().decode()
    return rst


def make_acl_entry(username, password, permissions):
    """
    It concat `username`, `digest` and `permissions` to a string as acl entry.
    :param username: the name of zookeeper user.
    :param password: the password of zookeeper user.
    :param permissions: a string, a list or a tuple.
    each element in `permissions` is a char who should be included in `cdrwa`. If `permissions` contains extra element, `PermTypeError` will be raised.
    :return: a string in form:
    "digest:<username>:<digest>:<permissions>"
    where `digest` is a string build by `zkutil.make_digest()`
    """
    perms = ""
    for c in permissions:
        if c not in PERM_TO_LONG:
            raise PermTypeError(c)
        perms += c

    return "digest:{username}:{digest}:{permissions}".format(
        username=username, digest=make_digest(username + ":" + password), permissions=perms
    )


def perm_to_long(short, lower=True):
    """
    Convert short style zookeeper permissions(`cdrwa` or `CDRWA`)
    to standard permission list(`['create', 'delete', 'read', 'write', 'admin']`).
    :param short: is `iterable` of short permissions that can be used in `for-loop`.
    Such as `"cdrw"` or `['c', 'd']`
    :param lower: specifies if convert result to lower or upper case.
    By default it is `True`, for lower case.
    :return: a list of standard permission.
    """
    rst = []

    for c in short:
        c = c.lower()
        if c not in PERM_TO_LONG:
            raise PermTypeError(c)

        rst.append(PERM_TO_LONG[c])

    if not lower:
        rst = [x.upper() for x in rst]

    return rst


def perm_to_short(lst, lower=True):
    """
    The reverse of `perm_to_long`:
    It convert a list of standard permissions back to a short permission string,
    such as `cdrw`.
    :param lst: is a list of standard permissions, such as `['create', 'read']`.
    :param lower: specifies if convert result to lower or upper case.
    By default it is `True`, for lower case.
    :return: a string of short permissions.
    """
    rst = ""

    for p in lst:
        p = p.lower()
        if p not in PERM_TO_SHORT:
            raise PermTypeError(p)

        rst += PERM_TO_SHORT[p]

    if not lower:
        rst = rst.upper()

    return rst


def make_kazoo_digest_acl(acl):
    # acl = (('xp', '123', 'cdrwa'),
    #        ('foo', 'passw', 'rw'),
    # )
    # print zkutil.make_kazoo_digest_acl(acl)
    #
    # [
    #     ACL(perms=12, acl_list=['CREATE', 'DELETE'],
    #         id=Id(scheme='digest', id=u'foo:VNy+Z9IdXrOUk9Rtia4fQS071t4=')),
    #     ACL(perms=31, acl_list=['ALL'],
    #         id=Id(scheme='digest', id=u'xp:LNcZO17uqqJ4GuYoSclIsGjYniQ='))
    # ]

    """

    :param acl: acl in tuple or list.
    :return: a `list` of `kazoo.security.ACL`.
    """
    if acl is None:
        return None

    rst = []
    for name, passw, perms in acl:
        perm_dict = {p: True for p in perm_to_long(perms)}
        acl_entry = security.make_digest_acl(name, passw, **perm_dict)
        rst.append(acl_entry)

    return rst


def parse_kazoo_acl(acls):
    """
    Convert kazoo style acls in list/tuple to a list in form `[(<scheme>, <user>, <perm>)]`.
    :param acls: a list/tuple of `kazoo.security.ACL`.
    :return: a list of acl in form `[(<scheme>, <user>, <perm>)]`
    """
    # acls = [ACL(perms=31,
    #            acl_list=['ALL'],
    #            id=Id(scheme='digest', id=u'user:+Ir5sN1lGJEEs8xBZhZXK='))]

    rst = []
    for acl in acls:
        if "ALL" in acl.acl_list:
            acl_list = "cdrwa"
        else:
            acl_list = perm_to_short(acl.acl_list)

        rst.append((acl.id.scheme, acl.id.id.split(":")[0], acl_list))

    return rst


def is_backward_locking(locked_keys, key):
    # ### Naive dead lock detect:
    #
    # Locks must be acquired in alphabetic order(or other order, from left to right.
    #     Trying to acquire a `lock>=right_most_locked`, is a forward locking Otherwise it
    #                                                                                   is a backward locking.
    #
    #     Always do forward locking we can guarantee there won't be a dead lock.
    # Since a deadlock needs at least one backward locking to form a circular dependency.
    #
    #     If a process fails to acquire a lock in a backward locking,
    #     it should release all locks it holds and redo the entire transaction.
    #
    #     E.g. suppose X has acquired lock a and c, Y has acquired lock b:
    #
    # ```
    # Acquired locks by process X and Y
    # locks are ordered left-to-right
    #                   ---------------------------------------------
    # proc-X   a(locked)                  c(locked)
    # proc-Y               b(locked)
    # ```
    #
    # If
    # X tries to acquire b(**backward**),
    #                    Y tries to acquire c(forward):
    # There is a deadlock. X should release all locks.
    #
    # ```
    # ---------------------------------------------
    # proc-X   a(locked)   .------------ c(locked)
    # v             ^
    # proc-Y               b(locked) ----'
    # ```
    #
    # If
    # X tries to acquire b(**backward**)
    # Y tries to acquires a(**backward**)
    # There is a deadlock, X and Y should both release their locks.
    #
    # ```
    # ---------------------------------------------
    # proc-X   a(locked)   .------------ c(locked)
    #          ^           v
    # proc-Y   '---------- b(locked)
    """
    Check if the operation of locking `key` is a backward-locking.
    :param locked_keys: is a collection support `in` operator that contains already locked keys.
    :param key: is the key to lock.
    :return: a `bool` indicate if locking `key` would be a backward-locking.
    """
    locked_keys = sorted(locked_keys)
    assert key not in locked_keys, "must not re-lock a key"

    if len(locked_keys) == 0:
        is_backward = False
    else:
        is_backward = key < locked_keys[-1]

    return is_backward


def _init_node(zkcli, parent_path, node, val, acl, users):
    path = parent_path + "/" + node

    if acl is None:
        acls = zkcli.get_acls(parent_path)[0]
    else:
        acls = [(user, users[user], perms) for user, perms in acl.items()]
        acls = make_kazoo_digest_acl(acls)

    if zkcli.exists(path) is None:
        zkcli.create(path, value=val, acl=acls)
    else:
        zkcli.set_acls(path, acls)

    return path


def init_hierarchy(hosts, hierarchy, users, auth):
    """
    It initialize a zookeeper cluster, including initializing the tree structure,
    setting value and permissions for each node.
    :param hosts: comma-separated list of hosts to connect to, such as `'127.0.0.1:2181,127.0.0.1:2182,[::1]:2183'`.
    :param hierarchy: a dict of zk node structure with value or acl optional for each node.
    For example,
    """
    # node1:
    #   __val__: "json, optional, by default a '{}'"
    #   __acl__: # optional, same with parent if not given
    #     user_1: "cdrwa"
    #     user_2: "rw"
    #                 ...
    # node2:
    #     node21: {...}
    """
    As shown above, each node has two attributes `__val__` and `__acl__` which are used to set the corresponding node.
    :param users:  a dict in form `{<username>: <password>}` containing all users in `hierarchy`.
    :param auth: a tuple in form `(<scheme>, <user>, <password>)`.
    It is the authorization info to connect to zookeeper which is used to initialize the zookeeper cluster.
    :return: None.
    """
    zkcli = KazooClient(hosts)
    zkcli.start()

    scheme, name, passw = auth
    zkcli.add_auth(scheme, name + ":" + passw)

    def _init_hierarchy(hierarchy, parent_path):
        if len(hierarchy) == 0:
            return

        for node, attr_children in hierarchy.items():
            val = attr_children.get("__val__", {})
            val = k3utfjson.dump(val).encode("utf-8")
            acl = attr_children.get("__acl__")

            path = _init_node(zkcli, parent_path, node, val, acl, users)
            children = {k: v for k, v in attr_children.items() if k not in ("__val__", "__acl__")}

            _init_hierarchy(children, path)

    _init_hierarchy(hierarchy, "/")
    close_zk(zkcli)


def _make_zk_path(*paths):
    return "/" + "/".join([x.strip("/") for x in paths])


def export_hierarchy(zkcli, zkpath):
    """
    Exporting a zookeeper node in a tree structure, and you can import the data into zookeeper
    using `zkutil.init_hierarchy`
    :param zkcli: is a `KazooClient` instance connected to zk.
    :param zkpath: is zookeeper root path that you want export
    :return:
    """
    if zkpath != "/":
        zkpath = zkpath.rstrip("/")

    if not zkpath.startswith("/"):
        raise ZkPathError("zkpath: {0} Error, Should be absolute path".format(zkpath))

    zk_node = _export_hierarchy(zkcli, zkpath)

    return zk_node


def _export_hierarchy(zkcli, zkpath):
    acls = {}

    value, stat = zkcli.get(zkpath)

    _acls, stat = zkcli.get_acls(zkpath)
    _acls = parse_kazoo_acl(_acls)

    for schema, user, perm in _acls:
        acls.update({user: perm})

    _zk_node = {"__val__": value, "__acl__": acls}

    for child in zkcli.get_children(zkpath):
        _zkpath = _make_zk_path(zkpath, child)

        _zk_node[child] = _export_hierarchy(zkcli, _zkpath)

    return _zk_node


NeedWait = namedtuple("NeedWait", [])


# A collection of helper functions those block until a condition satisfied before
# returning.
# Such as:
#
# -   Wait for a zk-node to be absent.
# -   Wait for a zk-node version greater than a specified one before returning
# the node value.
#
# These methods share the same form: `def xxx(zkclient, path, timeout=None, **kwargs)`.
#
# Common parameters are:
#
# -   `zkclient`:
# kazoo client.
#
# -   `path`:
# specifies the path to watch.
#
# -   `timeout`:
# specifies the time(in second) to wait.
#
# By default it is `None` which means to wait for a year.
#
# If condition was not satisfied in `timeout` seconds,
# it raises a `ZKWaitTimeout` exception.
def _conditioned_get_loop(zkclient, path, conditioned_get, timeout=None, **kwargs):
    if timeout is None:
        timeout = 86400 * 365

    expire_at = time.time() + timeout
    lck = threading.RLock()
    maybe_available = threading.Event()
    maybe_available.clear()

    def set_available():
        with lck:
            maybe_available.set()

    def on_connection_change(state):
        # notify it to re-get, then raise Connection related error
        logger.info("connection state change: {0}".format(state))
        set_available()

    zkclient.add_listener(on_connection_change)

    try:
        while True:
            it = conditioned_get(zkclient, path, **kwargs)
            next(it)
            rst = it.send((NeedWait, set_available))

            if rst is not NeedWait:
                return rst

            if maybe_available.wait(expire_at - time.time()):
                with lck:
                    maybe_available.clear()
                continue

            raise ZKWaitTimeout(
                "timeout({timeout} sec) waiting for {path} to satisfy: {cond}".format(
                    timeout=timeout, path=path, cond=str(kwargs)
                )
            )
    finally:
        try:
            zkclient.remove_listener(on_connection_change)
        except Exception as e:
            logger.info(repr(e) + " while removing on_connection_change")


def make_conditioned_getter(conditioned_get):
    def _wrap(zkclient, path, timeout=None, **kwargs):
        return _conditioned_get_loop(zkclient, path, conditioned_get, timeout=timeout, **kwargs)

    _wrap.__name__ = conditioned_get.__name__

    return _wrap


@make_conditioned_getter
def get_next(zkclient, path, version=-1):
    """
    Wait until zk-node `path` version becomes greater than `version` then return
    node value and `zstat`.
    :param version: the version that `path` version must be greater than.
    :return: zk node value and `zstat`.
    """
    NeedWait, set_available = yield

    val, zstat = zkclient.get(path, watch=lambda watchevent: set_available())
    if zstat.version > version:
        yield val, zstat

    yield NeedWait


@make_conditioned_getter
def wait_absent(zkclient, path):
    """
    Wait at most `timeout` seconds for zk-node `path` to be absent.

    If `path` does not exist, it returns at once.
    :param zkclient:
    :param path:
    :return:
    """
    NeedWait, set_available = yield

    try:
        zkclient.get(path, watch=lambda watchevent: set_available())
    except NoNodeError as e:
        logger.info(repr(e) + " found, return")
        yield None

    yield NeedWait



#!/usr/bin/env python2
# coding: utf-8

import time
import unittest

from kazoo.exceptions import NoNodeError

import k3thread
import k3utdocker
import k3utfjson
import k3zkutil
from k3zkutil.test.helper import wait_for_zk

zk_test_name = "zk_test"
zk_test_tag = "zookeeper:3.9"


class TestCachedReader(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        k3utdocker.pull_image(zk_test_tag)

    def setUp(self):
        k3utdocker.create_network()
        k3utdocker.start_container(
            zk_test_name,
            zk_test_tag,
            port_bindings={
                2181: 21811,
            },
        )

        self.zk = wait_for_zk("127.0.0.1:21811")
        self.val = {"a": 1, "b": 2}
        self.zk.create("foo", k3utfjson.dump(self.val).encode("utf-8"))

    def tearDown(self):
        self.zk.stop()
        k3utdocker.remove_container(zk_test_name)

    def test_cache(self):
        c = k3zkutil.CachedReader(self.zk, "foo")
        self.assertDictEqual(self.val, c)

    def test_cb(self):
        latest = ["foo"]

        def cb(path, old, new):
            latest[0] = new

        k3zkutil.CachedReader(self.zk, "foo", callback=cb)

        for i in range(100):
            self.val["a"] += 1
            self.zk.set("foo", k3utfjson.dump(self.val).encode("utf-8"))

        time.sleep(1)
        self.assertEqual(self.val, latest[0])

    def test_update(self):
        c = k3zkutil.CachedReader(self.zk, "foo")
        self.assertDictEqual(self.val, c)

        cases = (
            {"a": 2},
            {"a": "a_v", "b": "b_v"},
            {"a": 3, "b": {"c": 4}, "d": {"e": {"e": "val"}}},
        )

        for case in cases:
            self.zk.set("foo", k3utfjson.dump(case).encode("utf-8"))
            time.sleep(0.5)
            self.assertDictEqual(case, c)

    def test_ex(self):
        try:
            k3zkutil.CachedReader(self.zk, "bar")
            self.assertFalse(True, "should raise a NoNodeError")
        except NoNodeError:
            pass

        c = k3zkutil.CachedReader(self.zk, "foo")
        self.assertRaises(k3zkutil.ZKWaitTimeout, c.watch, 1)

    def test_watch(self):
        val = {"a": 2}
        c = k3zkutil.CachedReader(self.zk, "foo")

        def _change_node():
            self.zk.set("foo", k3utfjson.dump(val).encode("utf-8"))

        k3thread.daemon(_change_node, after=1)
        self.assertEqual([self.val, val], c.watch())

        def _close():
            c.close()

        k3thread.daemon(_close, after=1)
        self.assertEqual(None, c.watch())



import time
import unittest

from kazoo.client import KazooClient
from kazoo.exceptions import ConnectionClosedError

import k3thread
import k3ut
import k3utdocker
import k3utfjson
import k3zkutil
from k3confloader import conf
from k3zkutil.test.helper import wait_for_zk

dd = k3ut.dd

zk_test_name = "zk_test"
zk_test_tag = "zookeeper:3.9"


zk_test_auth = ("digest", "xp", "123")
zk_test_acl = (("xp", "123", "cdrw"),)

# zookeeper docker env vars:
# https://hub.docker.com/_/zookeeper/
#
# Example stack.yml for zookeeper:
#
# version: '3.1'
# services:
#   zoo1:
#     image: zookeeper
#     restart: always
#     hostname: zoo1
#     ports:
#       - 2181:2181
#     environment:
#       ZOO_MY_ID: 1
#       ZOO_SERVERS: server.1=0.0.0.0:2888:3888 server.2=zoo2:2888:3888 server.3=zoo3:2888:3888
#   zoo2: ...
#   zoo3: ...
#
# ZOO_TICK_TIME          : Defaults to 2000. ZooKeeper's tickTime
# ZOO_INIT_LIMIT         : Defaults to 5. ZooKeeper's initLimit
# ZOO_SYNC_LIMIT         : Defaults to 2. ZooKeeper's syncLimit
# ZOO_MAX_CLIENT_CNXNS   : Defaults to 60. ZooKeeper's maxClientCnxns
# ZOO_STANDALONE_ENABLED : Defaults to false. Zookeeper's standaloneEnabled
# ZOO_MY_ID
# ZOO_SERVERS
#
# Where to store data
#       This image is configured with volumes at /data and /datalog to hold the Zookeeper in-memory database snapshots and the transaction log of updates to the database, respectively.


class TestZKLock(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        k3utdocker.pull_image(zk_test_tag)

    def setUp(self):
        conf.zk_acl = zk_test_acl
        conf.zk_auth = zk_test_auth

        self.counter = 0
        self.running = True

        k3utdocker.create_network()
        k3utdocker.start_container(
            zk_test_name,
            zk_test_tag,
            port_bindings={
                2181: 21811,
            },
        )

        self.zk = wait_for_zk("127.0.0.1:21811")
        scheme, name, passw = zk_test_auth
        self.zk.add_auth(scheme, name + ":" + passw)

        # create lock base dir
        acl = k3zkutil.make_kazoo_digest_acl(zk_test_acl)
        self.zk.create("lock/", acl=acl)

        self.lck = k3zkutil.ZKLock("foo_name", zkclient=self.zk)

    def tearDown(self):
        self.zk.stop()
        k3utdocker.remove_container(zk_test_name)

    def _on_conn_change(self, state):
        self.lsn_count += 1

    def test_bounded_listener(self):
        # ensure that adding a bounded listener(self.on_xxx) is ok

        self.lsn_count = 0

        self.zk.add_listener(self._on_conn_change)
        self.zk.add_listener(self._on_conn_change)

        self.zk.stop()

        self.assertEqual(1, self.lsn_count)

    def _loop_acquire(self, n, ident):
        zk = KazooClient(hosts="127.0.0.1:21811")
        zk.start()
        scheme, name, passw = zk_test_auth
        zk.add_auth(scheme, name + ":" + passw)

        for ii in range(n):
            lock = k3zkutil.ZKLock("foo_name", zkclient=zk)
            with lock:
                self.total += 1
                self.counter += 1

                self.assertTrue(self.counter == 1)

                time.sleep(0.01)
                self.counter -= 1

                dd(
                    "id={ident:0>2} n={ii:0>2} got and released lock: {holder}".format(
                        ident=ident, ii=ii, holder=lock.lock_holder
                    )
                )

        zk.stop()

    def test_concurrent(self):
        self.running = True
        self.total = 0
        n_repeat = 40
        n_thread = 5

        ths = []
        for ii in range(n_thread):
            t = k3thread.daemon(
                self._loop_acquire,
                args=(
                    n_repeat,
                    ii,
                ),
            )
            ths.append(t)

        for th in ths:
            th.join()

        self.running = False
        self.assertEqual(n_repeat * n_thread, self.total)

    def test_persistent(self):
        lock = k3zkutil.ZKLock("foo_name", ephemeral=False, on_lost=lambda: True)
        try:
            with lock:
                lock.zkclient.stop()
        except ConnectionClosedError:
            pass

        self.assertRaises(k3zkutil.LockTimeout, self.lck.acquire, timeout=0.2)

    def test_timeout(self):
        l1 = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)
        l2 = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)

        with l1:
            with k3ut.Timer() as t:
                self.assertRaises(k3zkutil.LockTimeout, l2.acquire, timeout=0.2)
                self.assertAlmostEqual(0.2, t.spent(), places=1)

            with k3ut.Timer() as t:
                self.assertRaises(k3zkutil.LockTimeout, l2.acquire, timeout=-1)
                self.assertAlmostEqual(0.0, t.spent(), delta=0.01)

        try:
            l2.acquire(timeout=-1)
        except k3zkutil.LockTimeout:
            self.fail("timeout<0 should could acquire")

    def test_lock_holder(self):
        a = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)
        b = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)

        with a:
            self.assertIsInstance(a.identifier, dict)
            self.assertIsNone(a.identifier["val"], None)

            self.assertEqual((a.identifier, 0), a.lock_holder)
            val, zstate = self.zk.get(a.lock_path)
            val = k3utfjson.load(val)

            self.assertEqual((val, zstate.version), a.lock_holder)

            locked, holder, ver = b.try_acquire()
            self.assertFalse(locked)
            self.assertEqual((a.identifier, 0), (holder, ver))
            self.assertEqual((val, zstate.version), (holder, ver))

    def test_watch_acquire(self):
        a = k3zkutil.ZKLock("foo", on_lost=lambda: True)
        b = k3zkutil.ZKLock("foo", on_lost=lambda: True)

        # no one locked

        n = 0
        for holder, ver in a.acquire_loop():
            n += 1
        self.assertEqual(0, n, "acquired directly")

        # watch node change

        it = b.acquire_loop()

        holder, ver = next(it)
        self.assertEqual((a.identifier, 0), (holder, ver))

        a.identifier["val"] = "xx"
        value = k3utfjson.dump(a.identifier).encode("utf-8")
        self.zk.set(a.lock_path, value)

        holder, ver = next(it)
        self.assertEqual(("xx", 1), (holder["val"], ver), "watched node change")

        a.release()
        try:
            holder, ver = next(it)
            self.fail("should not have next yield")
        except StopIteration:
            pass

        self.assertTrue(b.is_locked())

    def test_set_locked_key(self):
        l1 = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)

        with l1:
            val, zstate = self.zk.get(l1.lock_path)
            val = k3utfjson.load(val)
            self.assertEqual(val, l1.identifier)
            self.assertEqual((val, zstate.version), l1.lock_holder)

            self.assertIsInstance(val, dict)
            self.assertIsNone(val["val"], None)

            l1.set_lock_val("foo_val", zstate.version)

            val, zstate = self.zk.get(l1.lock_path)
            val = k3utfjson.load(val)
            self.assertEqual(val, l1.identifier)

            self.assertIsInstance(val, dict)
            self.assertEqual(val["val"], "foo_val")

    def test_try_lock(self):
        l1 = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)
        l2 = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)

        with l1:
            with k3ut.Timer() as t:
                locked, holder, ver = l2.try_acquire()
                self.assertFalse(locked)
                self.assertEqual(l1.identifier, holder)
                self.assertGreaterEqual(ver, 0)

                self.assertAlmostEqual(0.0, t.spent(), delta=0.05)

        with k3ut.Timer() as t:
            locked, holder, ver = l2.try_acquire()
            self.assertTrue(locked)
            self.assertEqual(l2.identifier, holder)
            self.assertEqual(ver, 0)

            self.assertAlmostEqual(0.0, t.spent(), delta=0.05)

    def test_try_release(self):
        l1 = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)
        l2 = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)

        released, holder, ver = l1.try_release()
        self.assertEqual((True, l1.identifier, -1), (released, holder, ver))

        with l2:
            released, holder, ver = l1.try_release()
            self.assertEqual((False, l2.identifier, 0), (released, holder, ver))

            released, holder, ver = l2.try_release()
            self.assertEqual((True, l2.identifier, 0), (released, holder, ver))

    def test_zk_lost(self):
        sess = {"acquired": True}

        def watch(state):
            dd("zk node state changed to: ", state)
            sess["acquired"] = False

        self.zk.add_listener(watch)

        # test zk close

        lock = k3zkutil.ZKLock("foo_name", zkclient=self.zk)

        lock.acquire()
        self.zk.stop()
        time.sleep(0.1)
        self.assertFalse(sess["acquired"])

        # test node delete

        sess["acquired"] = True
        self.zk.start()
        self.zk.add_listener(watch)

        lock = k3zkutil.ZKLock("foo_name", zkclient=self.zk)

        with lock:
            time.sleep(0.1)
            self.zk.delete(lock.zkconf.lock("foo_name"))
            time.sleep(0.1)
            self.assertFalse(sess["acquired"])

    def test_node_change_after_acquired(self):
        sess = {"acquired": True}

        def on_lost():
            sess["acquired"] = False

        lock = k3zkutil.ZKLock("foo_name", zkclient=self.zk, on_lost=on_lost)

        with lock:
            sess["acquired"] = True
            self.zk.delete(lock.zkconf.lock("foo_name"))
            time.sleep(0.1)
            self.assertFalse(sess["acquired"])

        lock = k3zkutil.ZKLock("foo_name", zkclient=self.zk, on_lost=on_lost)

        with lock:
            sess["acquired"] = True
            self.zk.set(lock.zkconf.lock("foo_name"), b"xxx")
            time.sleep(0.1)
            self.assertFalse(sess["acquired"])

    def test_node_change_after_released(self):
        sess = {"acquired": True}

        def on_lost():
            sess["acquired"] = False

        lock = k3zkutil.ZKLock("foo_name", zkclient=self.zk, on_lost=on_lost)

        with lock:
            sess["acquired"] = True

        time.sleep(0.1)
        self.assertTrue(sess["acquired"])

    def test_is_locked(self):
        lock = k3zkutil.ZKLock("foo_name", zkclient=self.zk)

        with lock:
            pass

        self.assertFalse(lock.is_locked())

        lock = k3zkutil.ZKLock("foo_name", zkclient=self.zk)
        lock.acquire()
        self.assertTrue(lock.is_locked())
        lock.try_release()
        self.assertFalse(lock.is_locked())

    def test_conn_lost_when_blocking_acquiring(self):
        l2 = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)

        th = k3thread.daemon(target=self.zk.stop, after=0.5)
        with l2:
            try:
                self.lck.acquire(timeout=1)
                self.fail("expected connection error")
            except ConnectionClosedError:
                pass

        th.join()

    def test_internal_zkclient(self):
        sess = {"acquired": True}

        def on_lost():
            sess["acquired"] = False

        # There must be a listener specified to watch connection issue
        self.assertRaises(ValueError, k3zkutil.ZKLock, "foo_name")

        lock = k3zkutil.ZKLock("foo_name", on_lost=on_lost)

        with lock:
            self.zk.delete(lock.zkconf.lock("foo_name"))
            time.sleep(0.1)
            self.assertFalse(sess["acquired"])

    def test_acl(self):
        with self.lck:
            acls, zstat = self.zk.get_acls(self.lck.lock_path)

        dd(acls)
        self.assertEqual(1, len(acls))

        acl = acls[0]
        expected = k3zkutil.perm_to_long(zk_test_acl[0][2], lower=False)

        self.assertEqual(set(expected), set(acl.acl_list))
        self.assertEqual("digest", acl.id.scheme)
        self.assertEqual(zk_test_acl[0][0], acl.id.id.split(":")[0])

    def test_config(self):
        old = (conf.zk_acl, conf.zk_auth, conf.zk_node_id)

        conf.zk_acl = (("foo", "bar", "cd"), ("xp", "123", "cdrwa"))

        conf.zk_auth = ("digest", "xp", "123")
        conf.zk_node_id = "abc"

        lock = k3zkutil.ZKLock("foo_name", on_lost=lambda: True)

        dd(lock.zkconf.acl())

        def _check_ac(ac):
            self.assertEqual("digest", ac.id.scheme)
            self.assertEqual("foo", ac.id.id.split(":")[0])
            self.assertEqual(set(["CREATE", "DELETE"]), set(ac.acl_list))

        _check_ac(lock.zkconf.kazoo_digest_acl()[0])

        with lock:
            # should have created lock node
            data, zstate = self.zk.get(lock.lock_path)
            data = k3utfjson.load(data)["id"]
            dd(data)

            self.assertEqual("abc", data.split("-")[0])

            acls, zstate = self.zk.get_acls(lock.lock_path)
            dd(acls)

            _check_ac(acls[0])

        (conf.zk_acl, conf.zk_auth, conf.zk_node_id) = old

    def test_hosts(self):
        lock = k3zkutil.ZKLock(
            "foo_name",
            zkconf=dict(
                hosts="127.0.0.1:21811",
            ),
            on_lost=lambda: True,
        )

        with lock:
            self.assertEqual("127.0.0.1:21811", lock._hosts)

    def test_specify_identifier(self):
        a = k3zkutil.ZKLock(
            "foo_name",
            zkconf=dict(
                hosts="127.0.0.1:21811",
            ),
            identifier="faked",
            on_lost=lambda: True,
        )

        b = k3zkutil.ZKLock(
            "foo_name",
            zkconf=dict(
                hosts="127.0.0.1:21811",
            ),
            identifier="faked",
            on_lost=lambda: True,
        )

        a.acquire()
        b.acquire()
        dd("a and b has the same identifier thus they both can acquire the lock")

    def test_release_listener_removed(self):
        self.lck.release()
        self.assertNotIn(self.lck.on_connection_change, self.zk.state_listeners)

    def test_release_owning_client_stopped(self):
        lock = k3zkutil.ZKLock(
            "foo_name",
            zkconf=dict(
                hosts="127.0.0.1:21811",
            ),
            on_lost=lambda: True,
        )

        lock.release()
        self.assertTrue(lock.zkclient._stopped.is_set())



import unittest


import k3thread
import k3ut
import k3utdocker
import k3zkutil
from k3zkutil.test.helper import wait_for_zk

dd = k3ut.dd

zk_tag = "zookeeper:3.9"
zk_name = "zk_test"


class TestAcid(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        k3utdocker.pull_image(zk_tag)

    def setUp(self):
        k3utdocker.create_network()
        k3utdocker.start_container(
            zk_name,
            zk_tag,
            port_bindings={2181: 21811},
        )

        self.zk = wait_for_zk("127.0.0.1:21811")

        dd("start zk-test in docker")

        self.path = "a"
        self.zk.create(self.path, b"1")

    def tearDown(self):
        self.zk.stop()
        k3utdocker.remove_container(zk_name)

    def test_cas(self):
        for curr in k3zkutil.cas_loop(self.zk, self.path):
            curr.v += 2

        final_val, zstat = self.zk.get(self.path)
        dd(final_val, zstat)
        self.assertEqual(b"3", final_val)

    def test_cas_abort(self):
        for curr in k3zkutil.cas_loop(self.zk, self.path):
            curr.v += 2
            break

        final_val, zstat = self.zk.get(self.path)
        dd(final_val, zstat)
        self.assertEqual(b"1", final_val, "a break statement cancel set_val()")

    def test_cas_create_zk(self):
        for first_arg in (
            "127.0.0.1:21811",
            {"hosts": "127.0.0.1:21811"},
            k3zkutil.ZKConf(hosts="127.0.0.1:21811"),
        ):
            self.zk.set(self.path, b"1")

            for curr in k3zkutil.cas_loop(first_arg, self.path):
                curr.v += 2

            final_val, zstat = self.zk.get(self.path)
            dd(final_val, zstat)
            self.assertEqual(b"3", final_val)

    def test_cas_non_json(self):
        for curr in k3zkutil.cas_loop(self.zk, self.path, json=False):
            self.assertIsInstance(curr.v, bytes)
            curr.v += b"a"

        final_val, zstat = self.zk.get(self.path)
        dd(final_val, zstat)
        self.assertEqual(b"1a", final_val)

    def test_cas_concurrent(self):
        def _update():
            for ii in range(10):
                for curr in k3zkutil.cas_loop("127.0.0.1:21811", self.path):
                    curr.v += 1

        ths = [k3thread.daemon(_update) for _ in range(5)]

        for th in ths:
            th.join()

        final_val, zstat = self.zk.get(self.path)
        dd(final_val, zstat)

        self.assertEqual(b"51", final_val)



import k3zkutil

"""
config.zk_acl      # (('xp', '123', 'cdrwa'), ('foo', 'bar', 'rw'))
config.zk_auth     # ('digest', 'xp', '123')
config.zk_hosts    # '127.0.0.1:2181'
config.zk_node_id  # 'web-01'
config.zk_lock_dir # 'lock/'
"""
with k3zkutil.ZKLock(
    "foo_lock",
    zkconf=dict(
        hosts="127.0.0.1:2181",
        acl=(("xp", "123", "cdrwa"),),
        auth=("digest", "xp", "123"),
        node_id="web-3",
        lock_dir="my_locks/",
    ),
):
    print("do something")
lock = k3zkutil.ZKLock("foo")
try:
    for holder, ver in lock.acquire_loop(timeout=3):
        print("lock is currently held by:", holder, ver)

    print("lock is acquired")
except k3zkutil.LockTimeout:
    print('timeout to acquire "foo"')



#!/usr/bin/env python2
# coding: utf-8

import logging
import threading

from . import zkconf
from . import zkutil
from . import exceptions

logger = logging.getLogger(__name__)


class CachedReader(dict):
    # bar = {
    #    'jobs': {
    #        'num' : 10
    #    }
    # }
    # cr = CachedReader('127.0.0.1:2181', 'bar')
    # for i in range(cr['jobs']['num']):
    #     doit()
    def __init__(self, zk, path, callback=None):
        """

        :param zk: is the connection argument, which can be:
        -   Comma separated host list, such as
            `"127.0.0.1:2181,127.0.0.2:2181"`.

        -   A `zkutil.ZKConf` instance specifying connection argument with other
            config.

        -   A plain `dict` to create a `zkutil.ZKConf` instance.
        :param path: the path of the node in zookeeper.
        :param callback: give a callback when the node change. Defaults to `None`.
        It has 3 arguments `(path, old_dict, new_dict)`.
        """
        super(CachedReader, self).__init__()

        self.zke, self.owning_zk = zkconf.kazoo_client_ext(zk)
        self.path = path
        self.callback = callback
        self.available_ev = threading.Event()
        self.stopped = False
        self.val = [None, None]
        # lock for update the dict
        self.lock = threading.RLock()

        def _conn_change_cb(state):
            self._on_conn_change(state)

        def _node_change_cb(event):
            self._on_node_change(event)

        self.conn_change_cb = _conn_change_cb
        self.node_change_cb = _node_change_cb
        self.zke.add_listener(self.conn_change_cb)
        self._update()

    def watch(self, timeout=None):
        """
        Wait until the node change and return a list `[old_dict, new_dict]`.
        If timeout, raise a `ZKWaitTimeout`.
        :param timeout: specifies the time(in second) to wait.
        By default it is `None` which means to wait for a year
        :return: If close the `CachedReader` by `zkutil.CachedReader.close()`, it return `None`.
        If the node change, it return a list `[old_dict, new_dict]`
        """
        self.available_ev.clear()

        timeout = timeout or 86400 * 365

        if self.available_ev.wait(timeout):
            if self.stopped:
                return None

            else:
                return self.val

        else:
            raise exceptions.ZKWaitTimeout("timeout {t} sec".format(t=timeout))

    def close(self):
        """
        Stop the `zkutil.CachedReader.watch` and the callback.
        :return: nothing
        """
        self.stopped = True
        self.available_ev.set()

        self.zke.remove_listener(self.conn_change_cb)
        if self.owning_zk:
            zkutil.close_zk(self.zke)

    def _on_conn_change(self, state):
        logger.info("state changed: {state}".format(state=state))
        self.stopped = True
        self.available_ev.set()

    def _on_node_change(self, event):
        logger.info("node state changed:{ev}".format(ev=event))

        if self.stopped:
            return

        self._update()
        self.available_ev.set()

        if self.callback is None:
            return

        self.callback(self.path, self.val[0], self.val[1])

    def _update(self):
        with self.lock:
            curr, _ = self.zke.get(self.path, watch=self.node_change_cb)
            self.val = [self.val[1], curr]
            self.update(curr)

            keys = list(self.keys())
            for k in keys:
                if k not in curr:
                    del self[k]



#!/usr/bin/env python
# coding: utf-8

import logging
import threading
import time

import k3utfjson

from kazoo.client import KazooClient
from kazoo.exceptions import LockTimeout
from kazoo.exceptions import NodeExistsError
from kazoo.exceptions import NoNodeError
from .exceptions import ZKUtilError

from . import zkutil
from .zkconf import ZKConf
from .zkconf import KazooClientExt

logger = logging.getLogger(__name__)


class ZKTXError(Exception):
    pass


class ZKTXConnectionLost(ZKTXError):
    pass


# TODO add method: ZKLock.get_owner() to get current lock owner
class ZKLock(object):
    """
    ZKLock implements a zookeeper based distributed lock.
    """

    def __init__(
        self, lock_name, zkconf=None, zkclient=None, on_lost=None, identifier=None, ephemeral=True, timeout=10
    ):
        if zkconf is None:
            zkconf = ZKConf()
        if isinstance(zkconf, dict):
            zkconf = ZKConf(**zkconf)
        self.zkconf = zkconf

        if zkclient is None:
            # If user does not pass a zkclient instance,
            # we need to create one for lock to use.
            # This zkclient will be closed after lock is released

            if on_lost is None:
                raise ValueError("on_lost must be specified to watch zk connection issue if no zkclient specified")

            zkclient = make_owning_zkclient(zkconf.hosts(), zkconf.auth())
            self.owning_client = True
        else:
            self.owning_client = False

        if identifier is None:
            identifier = zkutil.lock_id(zkconf.node_id())

        if not isinstance(identifier, dict):
            identifier = make_identifier(identifier, None)

        assert sorted(["id", "val"]) == sorted(list(identifier.keys()))

        # a copy of hosts for debugging and tracking
        self._hosts = ",".join(["{0}:{1}".format(*x) for x in zkclient.hosts])

        self.zkclient = zkclient
        if isinstance(self.zkclient, KazooClientExt):
            self.zkclient = self.zkclient._zk

        self.on_lost = on_lost

        self.lock_name = lock_name
        self.lock_path = zkconf.lock(self.lock_name)
        self.identifier = identifier
        self.ephemeral = ephemeral
        self.timeout = timeout

        self.mutex = threading.RLock()
        self.maybe_available = threading.Event()
        self.maybe_available.set()
        self.lock_holder = None

        logger.info("adding event listener: {s}".format(s=self))
        self.zkclient.add_listener(self.on_connection_change)

    def on_node_change(self, watchevent):
        # Must be locked first.
        # Or there is a chance on_node_change is triggered before
        #         self.maybe_available.clear()
        with self.mutex:
            self.maybe_available.set()

            # If locked. the node change is treated as losing a lock
            if self.is_locked():
                if self.on_lost is not None:
                    self.on_lost()

        logger.info("node state changed:{ev}, lock might be released: {s}".format(ev=watchevent, s=str(self)))

    def on_connection_change(self, state):
        # notify zklock to re-do acquiring procedure, to trigger Connection Error
        with self.mutex:
            self.maybe_available.set()

        if self.on_lost is not None:
            self.on_lost()

    def acquire_loop(self, timeout=None):
        if timeout is None:
            timeout = self.timeout

        expire_at = time.time() + timeout

        while True:
            # Even if timeout is smaller than 0, try-loop continue on until
            # maybe_available is not ready.
            #
            # There is a chance that:
            #  - Failed to create lock node(lock is occupied by other)
            #  - Failed to get lock node(just deleted)
            #  - Failed to create lock node(lock is occupied by other)
            #  - Failed to get lock node(just deleted)
            #  - ...
            if not self.maybe_available.wait(timeout=expire_at - time.time()):
                logger.debug("lock is still held by others: " + str(self))

                if time.time() > expire_at:
                    raise LockTimeout("lock: " + str(self.lock_path))

            # Always proceed the "get" phase, in order to add a watch handler.
            # To watch node change event.

            self._create()
            self._acquire_by_get()
            if self.is_locked():
                return

            # If it is possible to acquire the lock in next retry, do not yield
            if self.maybe_available.is_set():
                continue
            else:
                yield self.lock_holder[0], self.lock_holder[1]

    def acquire(self, timeout=None):
        for holder, ver in self.acquire_loop(timeout=timeout):
            continue

    def try_acquire(self):
        """
        Try to acquire the lock and return result.
        It never blocks.
        :return: a tuple of result, lock holder and lock holder version.
        Such as `(True, "aa-xx-bb", -1)` or `(False, "aa-xx-cc", 12)`
        """
        # If lock is acquired:
        # - the 1st element is `True`,
        # - the 2nd is identifier of this lock,
        # - the 3rd is `-1`.
        #
        # If lock is not acquired:
        # - the 1st element is `False`,
        # - the 2nd is identifier of the lock holder,
        # - the 3rd is a non-negative integer, which is the version of the zk node.
        try:
            self.acquire(timeout=-1)
        except LockTimeout:
            pass

        # if_locked, lock holder identifier, holder version
        return self.is_locked(), self.lock_holder[0], self.lock_holder[1]

    def try_release(self):
        """
        Release lock if current lock holder is this lock.
        It never blocks.
        :return: a tuple of result, lock holder and lock holder version.
        Such as `(True, "aa-xx-bb", -1)` or `(False, "aa-xx-cc", 12)`
        """
        # If lock holder is this `ZKLock`(checked by identifier), or lock is not acquired
        # by anyone):
        #
        # - the 1st element is `True`,
        #           - the 2nd is identifier of this lock,
        #                                           - the 3rd is `-1` or `zstat.version` or the lock zknode.
        #
        # Otherwise:
        #
        # - the 1st element is `False`,
        #           - the 2nd is identifier of the lock holder,
        #           - the 3rd is a non-negative integer, which is the version of the zk node.
        logger.debug("try to release if I am locker holder: {s}".format(s=str(self)))

        try:
            holder, zstat = self.zkclient.get(self.lock_path)
            holder = k3utfjson.load(holder)

            self.lock_holder = (holder, zstat.version)

            logger.debug("got lock holder: {s}".format(s=str(self)))

            if self.cmp_identifier(holder, self.identifier):
                self.zkclient.remove_listener(self.on_connection_change)

                try:
                    self.zkclient.delete(self.lock_path, version=zstat.version)
                except NoNodeError as e:
                    logger.info(repr(e) + " while delete lock: " + str(self))

                self.lock_holder = None

                return True, holder, zstat.version
            else:
                return False, holder, zstat.version

        except NoNodeError as e:
            logger.info(repr(e) + " while try_release: {s}".format(s=str(self)))
            return True, self.identifier, -1

    def release(self):
        """
        Release the lock if it has been locked.
        Otherwise return silently.

        If this lock initiated a connection by itself, it will be closed.

        :return: Nothing
        """
        with self.mutex:
            if self.is_locked():
                # remove listener to avoid useless event triggering
                self.zkclient.remove_listener(self.on_connection_change)

                try:
                    self.zkclient.delete(self.lock_path)
                except NoNodeError as e:
                    logger.info(repr(e) + " while delete lock: " + str(self))

                self.lock_holder = None

                logger.info("RELEASED: {s}".format(s=str(self)))
            else:
                logger.info("not acquired, do not need to release")

        self.close()

    def close(self):
        self.zkclient.remove_listener(self.on_connection_change)

        if self.owning_client:
            logger.info("zk client is made by me, close it")
            zkutil.close_zk(self.zkclient)

    def is_locked(self):
        holder = self.lock_holder
        if holder is None:
            return False

        return self.cmp_identifier(holder[0], self.identifier)

    def _create(self):
        logger.debug("to creaet: {s}".format(s=str(self)))

        try:
            self.zkclient.create(
                self.lock_path,
                k3utfjson.dump(self.identifier).encode("utf-8"),
                ephemeral=self.ephemeral,
                acl=self.zkconf.kazoo_digest_acl(),
            )

        except NodeExistsError as e:
            # NOTE Success create on server side might also results in failure
            # on client side due to network issue.
            # 'get' after 'create' to check if existent node belongs to this
            # client.

            logger.debug(repr(e) + " while create lock: {s}".format(s=str(self)))
            self.lock_holder = None
            return

        logger.info("CREATE OK: {s}".format(s=str(self)))

    def set_lock_val(self, val, version=-1):
        locked, holder, ver = self.try_acquire()
        if not locked:
            raise ZKUtilError("set non-locked: {k}".format(k=self.lock_name))

        self.identifier["val"] = val

        value = k3utfjson.dump(self.identifier).encode("utf-8")
        st = self.zkclient.set(self.lock_path, value, version=version)

        return st.version

    def get_lock_val(self):
        holder, zstat = self.zkclient.get(self.lock_path)
        holder = k3utfjson.load(holder)

        return holder["val"], zstat.version

    def cmp_identifier(self, ia, ib):
        return ia["id"] == ib["id"]

    def _acquire_by_get(self):
        logger.debug("to get: {s}".format(s=str(self)))

        try:
            with self.mutex:
                holder, zstat = self.zkclient.get(self.lock_path, watch=self.on_node_change)
                holder = k3utfjson.load(holder)

                self.lock_holder = (holder, zstat.version)

                logger.debug("got lock holder: {s}".format(s=str(self)))

                if self.cmp_identifier(holder, self.identifier):
                    logger.info("ACQUIRED: {s}".format(s=str(self)))
                    return

                logger.debug("other holds: {s}".format(s=str(self)))
                self.maybe_available.clear()

        except NoNodeError as e:
            # create failed but when getting it, it has been deleted
            logger.info(repr(e) + " while get lock: {s}".format(s=str(self)))
            with self.mutex:
                self.lock_holder = None
                self.maybe_available.set()

    def __str__(self):
        return "<id={id} {l}:[{holder}] on {h}>".format(
            id=self.identifier["id"],
            l=self.lock_path,
            holder=(self.lock_holder or ""),
            h=str(self._hosts),
        )

    def __enter__(self):
        self.acquire()

    def __exit__(self, tp, value, tb):
        self.release()


def make_owning_zkclient(hosts, auth):
    zkclient = KazooClient(hosts=hosts)
    zkclient.start()

    if auth is not None:
        scheme, name, passw = auth
        zkclient.add_auth(scheme, name + ":" + passw)

    return zkclient


def make_identifier(_id, val):
    return {"id": _id, "val": val}



import os
import time
import unittest
import uuid

from kazoo import security
from kazoo.client import KazooClient
from kazoo.exceptions import ConnectionClosedError
from kazoo.exceptions import NoNodeError
from k3confloader import conf

import k3net
import k3thread
import k3ut
import k3utdocker
import k3zkutil
from k3zkutil.test.helper import wait_for_zk

dd = k3ut.dd

zk_tag = "zookeeper:3.9"
zk_name = "zk_test"


class Testk3zkutil(unittest.TestCase):
    def test_lock_id(self):
        k = k3zkutil.lock_id("a")
        dd(k)
        elts = k.split("-")

        self.assertEqual(4, len(elts))

        self.assertEqual("a", elts[0])
        self.assertTrue(k3net.is_ip4(elts[1]))
        self.assertEqual(os.getpid(), int(elts[2]))

    def test_lock_id_default(self):
        expected = "%012x" % uuid.getnode()

        k = k3zkutil.lock_id()
        dd(conf)
        dd(k)
        self.assertEqual(expected, k.split("-")[0])

        k = k3zkutil.lock_id(node_id=None)
        dd(k)
        self.assertEqual(expected, k.split("-")[0])

        conf.zk_node_id = "a"
        k = k3zkutil.lock_id(node_id=None)
        dd(k)
        self.assertEqual("a", k.split("-")[0])

    def test_parse_lock_id(self):
        cases = (
            ("", ("", None, None)),
            ("-", ("", "", None)),
            ("--", ("", "", None)),
            ("1-", ("1", "", None)),
            ("1-a", ("1", "a", None)),
            ("1-a-x", ("1", "a", None)),
            ("1-a-1", ("1", "a", 1)),
        )

        for inp, expected in cases:
            rst = k3zkutil.parse_lock_id(inp)

            self.assertEqual(set(["node_id", "ip", "process_id", "uuid", "txid"]), set(rst.keys()))

            self.assertEqual(expected, (rst["node_id"], rst["ip"], rst["process_id"]))

    def test_parse_lock_id_with_txid(self):
        rst = k3zkutil.parse_lock_id("txid:123-a-b-x")
        self.assertEqual("txid:123", rst["node_id"])
        self.assertEqual("123", rst["txid"])

    def test_make_digest(self):
        cases = (("aa:", "E/ZoYMT80fFT7vhICWyvMdWNt7o="), ("username:password", "+Ir5sN1lGJEEs8xBZhZXKvjLJ7c="))

        for inp, expected in cases:
            rst = k3zkutil.make_digest(inp)
            self.assertEqual(expected, rst)

    def test_make_acl_entry(self):
        username = "zookeeper_user"
        password = "MT80fFT7vh"
        perm_cases = [
            "",
            "rw",
            "cdrwa",
            [],
            ["c"],
            ["c", "r", "r"],
            (),
            ("c",),
            ("c", "r", "w"),
            iter("cdrwa"),
        ]

        for inp in perm_cases:
            if isinstance(inp, type(iter(""))):
                perm = "cdrwa"
            else:
                perm = "".join(inp)

            dd("inp=", inp)
            dd("perm=", perm)
            rst = k3zkutil.make_acl_entry(username, password, inp)

            rst_splited = rst.split(":")
            self.assertEqual(4, len(rst_splited))
            self.assertEqual(("digest", username, perm), (rst_splited[0], rst_splited[1], rst_splited[3]))

        cases = (
            ((username, password, ""), "digest:zookeeper_user:Ds8aM7UNwfAlTN3IRdkBoCno9FM=:"),
            (("aabb", "abc123", "cdrwa"), "digest:aabb:t9MeAsoEPfdQEQjqbWtw8EHD9T0=:cdrwa"),
        )

        for inp, expected in cases:
            rst = k3zkutil.make_acl_entry(inp[0], inp[1], inp[2])
            self.assertEqual(expected, rst)

        invalid_perm_cases = (
            "abc",
            ["cde"],
            ["a", "v"],
            ("rw",),
            ("a", "b", "c"),
        )

        for inp in invalid_perm_cases:
            with self.assertRaises(k3zkutil.PermTypeError):
                k3zkutil.make_acl_entry(username, password, inp)

    def test_permission_convert(self):
        perm_cases = (
            ("", [], ""),
            ("rw", ["read", "write"], "rw"),
            ("cdrwa", ["create", "delete", "read", "write", "admin"], "cdrwa"),
            ([], [], ""),
            (["c"], ["create"], "c"),
            (["c", "r", "r"], ["create", "read", "read"], "crr"),
            ((), [], ""),
            (("c",), ["create"], "c"),
            (("c", "r", "w"), ["create", "read", "write"], "crw"),
            (iter("cdrwa"), ["create", "delete", "read", "write", "admin"], "cdrwa"),
        )

        for inp, lng, short in perm_cases:
            rst = k3zkutil.perm_to_long(inp)
            self.assertEqual(lng, rst)

            rst = k3zkutil.perm_to_short(lng)
            self.assertEqual(short, rst)

        # invalid short format

        invalid_short = (
            "abc",
            ["cde"],
            ["a", "v"],
            ("rw",),
            ("a", "b", "c"),
        )

        for inp in invalid_short:
            self.assertRaises(k3zkutil.PermTypeError, k3zkutil.perm_to_long, inp)

        # invalid long format

        invalid_long = (
            "abc",
            ["foo"],
        )

        for inp in invalid_long:
            self.assertRaises(k3zkutil.PermTypeError, k3zkutil.perm_to_short, inp)

    def test_permission_convert_case(self):
        self.assertEqual(["CREATE", "DELETE", "READ", "WRITE", "ADMIN"], k3zkutil.perm_to_long("cdrwa", lower=False))

        self.assertEqual(["CREATE", "DELETE", "READ", "WRITE", "ADMIN"], k3zkutil.perm_to_long("CDRWA", lower=False))

        self.assertEqual(["create", "delete", "read", "write", "admin"], k3zkutil.perm_to_long("CDRWA"))

        self.assertEqual("CDRWA", k3zkutil.perm_to_short(["create", "delete", "read", "write", "admin"], lower=False))

        self.assertEqual("CDRWA", k3zkutil.perm_to_short(["CREATE", "DELETE", "READ", "WRITE", "ADMIN"], lower=False))
        self.assertEqual("cdrwa", k3zkutil.perm_to_short(["CREATE", "DELETE", "READ", "WRITE", "ADMIN"]))

    def test_make_kazoo_digest_acl(self):
        inp = (("foo", "bar", "cd"), ("xp", "123", "cdrwa"))

        dd(inp)

        rst = k3zkutil.make_kazoo_digest_acl(inp)
        dd(rst)

        self.assertEqual(2, len(rst))

        ac = rst[0]
        self.assertEqual("digest", ac.id.scheme)
        self.assertEqual("foo", ac.id.id.split(":")[0])
        self.assertEqual(set(["CREATE", "DELETE"]), set(ac.acl_list))

        ac = rst[1]
        self.assertEqual("digest", ac.id.scheme)
        self.assertEqual("xp", ac.id.id.split(":")[0])
        self.assertEqual(set(["ALL"]), set(ac.acl_list))

        self.assertIsNone(k3zkutil.make_kazoo_digest_acl(None))

    def test_parse_kazoo_acl(self):
        inp = (
            security.make_acl("world", "anyone", all=True),
            security.make_digest_acl("foo", "bar", create=True, read=True),
            security.make_digest_acl("xp", "123", all=True),
        )
        expected = (("world", "anyone", "cdrwa"), ("digest", "foo", "rc"), ("digest", "xp", "cdrwa"))

        dd(inp)

        rst = k3zkutil.parse_kazoo_acl(inp)
        dd(rst)

        self.assertEqual(expected, tuple(rst))

    def test_is_backward_locking(self):
        cases = (
            ([], "a", False, None),
            (["a"], "a", False, AssertionError),
            (["a", "c"], "a", True, AssertionError),
            (["a", "c"], "c", True, AssertionError),
            (["a", "c"], "", True, None),
            (["a", "c"], "b", True, None),
            (["a", "c"], "d", False, None),
        )

        for locked, key, expected, err in cases:
            if err is None:
                rst = k3zkutil.is_backward_locking(locked, key)
                self.assertEqual(expected, rst)
            else:
                self.assertRaises(err, k3zkutil.is_backward_locking, locked, key)


class TestZKinit(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        k3utdocker.pull_image(zk_tag)

    def setUp(self):
        k3utdocker.create_network()

        k3utdocker.start_container(zk_name, zk_tag, port_bindings={2181: 21811})

        # Wait for zookeeper to be ready
        zk = wait_for_zk("127.0.0.1:21811")
        zk.stop()

        dd("start zk-test in docker")

    def tearDown(self):
        k3utdocker.stop_container(zk_name)
        k3utdocker.remove_container(zk_name)

        dd("remove_container: " + zk_name)

    def test_init_hierarchy(self):
        auth = ("digest", "aa", "pw_aa")
        hosts = "127.0.0.1:21811"
        users = {"aa": "pw_aa", "bb": "pw_bb", "cc": "pw_cc"}
        hierarchy = {
            "node1": {
                "__val__": b"node1_val",
                "__acl__": {"aa": "cdrwa", "bb": "rw"},
                "node11": {
                    "__val__": b"node11_val",
                    "__acl__": {"aa": "cdrwa", "cc": "r"},
                },
                "node12": {"__val__": b"node12_val", "node121": {"__val__": "node121_val"}},
                "node13": {"__acl__": {"aa": "cdrwa"}},
            },
            "node2": {
                "__val__": b"node2_val",
                "node21": {"__val__": b"node21_val"},
                "node22": {"__acl__": {"aa": "rwa"}},
            },
            "node3": {"__acl__": {"aa": "carw", "cc": "r"}, "node31": {"node311": {"node3111": {}, "node3112": {}}}},
        }

        k3zkutil.init_hierarchy(hosts, hierarchy, users, auth)

        zkcli = KazooClient(hosts)
        zkcli.start()
        zkcli.add_auth("digest", "aa:pw_aa")

        expected_nodes = (
            (
                "/node1",
                b'"node1_val"',
                [("digest", "aa", "cdrwa"), ("digest", "bb", "rw")],
                set(["node11", "node12", "node13"]),
            ),
            ("/node1/node11", b'"node11_val"', [("digest", "aa", "cdrwa"), ("digest", "cc", "r")], set([])),
            ("/node1/node12", b'"node12_val"', [("digest", "aa", "cdrwa"), ("digest", "bb", "rw")], set(["node121"])),
            ("/node1/node12/node121", b'"node121_val"', [("digest", "aa", "cdrwa"), ("digest", "bb", "rw")], set([])),
            ("/node1/node13", b"{}", [("digest", "aa", "cdrwa")], set([])),
            ("/node2", b'"node2_val"', [("world", "anyone", "cdrwa")], set(["node21", "node22"])),
            ("/node2/node21", b'"node21_val"', [("world", "anyone", "cdrwa")], set([])),
            ("/node2/node22", b"{}", [("digest", "aa", "rwa")], set([])),
            ("/node3", b"{}", [("digest", "aa", "rwca"), ("digest", "cc", "r")], set(["node31"])),
            ("/node3/node31", b"{}", [("digest", "aa", "rwca"), ("digest", "cc", "r")], set(["node311"])),
            (
                "/node3/node31/node311",
                b"{}",
                [("digest", "aa", "rwca"), ("digest", "cc", "r")],
                set(["node3111", "node3112"]),
            ),
            ("/node3/node31/node311/node3111", b"{}", [("digest", "aa", "rwca"), ("digest", "cc", "r")], set([])),
            ("/node3/node31/node311/node3112", b"{}", [("digest", "aa", "rwca"), ("digest", "cc", "r")], set([])),
        )

        for node, val, acl, children in expected_nodes:
            actual_acl = k3zkutil.parse_kazoo_acl(zkcli.get_acls(node)[0])
            self.assertEqual(val, zkcli.get(node)[0])
            self.assertEqual(acl, actual_acl)
            self.assertEqual(children, set(zkcli.get_children(node)))

        zkcli.stop()

    def test_export_hierarchy(self):
        auth = ("digest", "aa", "pw_aa")
        hosts = "127.0.0.1:21811"
        users = {"aa": "pw_aa", "bb": "pw_bb", "cc": "pw_cc"}

        hierarchy = {
            "node0": {"__val__": b"", "__acl__": {"aa": "cdrwa", "bb": "rw"}},
            "node1": {
                "__val__": b"node1_val",
                "__acl__": {"aa": "cdrwa", "bb": "rw"},
                "node11": {
                    "__val__": b"node11_val",
                    "__acl__": {"aa": "cdrwa", "cc": "r"},
                },
                "node12": {"__val__": b"node12_val", "node121": {"__val__": "node121_val"}},
                "node13": {"__acl__": {"aa": "cdrwa"}},
            },
            "node2": {
                "__val__": b"node2_val",
                "node21": {"__val__": b"node21_val"},
                "node22": {"__acl__": {"aa": "rwa"}},
            },
            "node3": {"__acl__": {"aa": "carw", "cc": "r"}, "node31": {"node311": {"node3111": {}, "node3112": {}}}},
        }

        k3zkutil.init_hierarchy(hosts, hierarchy, users, auth)

        zkcli = KazooClient(hosts)
        zkcli.start()
        zkcli.add_auth("digest", "aa:pw_aa")

        invalid_zkpath_cases = ("a", "a/", "a/b")

        for zkpath in invalid_zkpath_cases:
            with self.assertRaises(k3zkutil.ZkPathError):
                k3zkutil.zkutil.export_hierarchy(zkcli, zkpath)

        valid_cases = (
            (
                "/",
                {
                    "__acl__": {"anyone": "cdrwa"},
                    "__val__": b"",
                    "node0": {"__acl__": {"aa": "cdrwa", "bb": "rw"}, "__val__": b'""'},
                    "node1": {
                        "__acl__": {"aa": "cdrwa", "bb": "rw"},
                        "__val__": b'"node1_val"',
                        "node11": {"__acl__": {"aa": "cdrwa", "cc": "r"}, "__val__": b'"node11_val"'},
                        "node12": {
                            "__acl__": {"aa": "cdrwa", "bb": "rw"},
                            "__val__": b'"node12_val"',
                            "node121": {"__acl__": {"aa": "cdrwa", "bb": "rw"}, "__val__": b'"node121_val"'},
                        },
                        "node13": {"__acl__": {"aa": "cdrwa"}, "__val__": b"{}"},
                    },
                    "node2": {
                        "__acl__": {"anyone": "cdrwa"},
                        "__val__": b'"node2_val"',
                        "node21": {"__acl__": {"anyone": "cdrwa"}, "__val__": b'"node21_val"'},
                        "node22": {"__acl__": {"aa": "rwa"}, "__val__": b"{}"},
                    },
                    "node3": {
                        "__acl__": {"aa": "rwca", "cc": "r"},
                        "__val__": b"{}",
                        "node31": {
                            "__acl__": {"aa": "rwca", "cc": "r"},
                            "__val__": b"{}",
                            "node311": {
                                "__acl__": {"aa": "rwca", "cc": "r"},
                                "__val__": b"{}",
                                "node3111": {"__acl__": {"aa": "rwca", "cc": "r"}, "__val__": b"{}"},
                                "node3112": {"__acl__": {"aa": "rwca", "cc": "r"}, "__val__": b"{}"},
                            },
                        },
                    },
                    "zookeeper": {
                        "__acl__": {"anyone": "cdrwa"},
                        "__val__": b"",
                        "quota": {"__acl__": {"anyone": "cdrwa"}, "__val__": b""},
                    },
                },
            ),
            ("/node0", {"__acl__": {"aa": "cdrwa", "bb": "rw"}, "__val__": b'""'}),
            (
                "/node1",
                {
                    "__acl__": {"aa": "cdrwa", "bb": "rw"},
                    "__val__": b'"node1_val"',
                    "node11": {"__acl__": {"aa": "cdrwa", "cc": "r"}, "__val__": b'"node11_val"'},
                    "node12": {
                        "__acl__": {"aa": "cdrwa", "bb": "rw"},
                        "__val__": b'"node12_val"',
                        "node121": {"__acl__": {"aa": "cdrwa", "bb": "rw"}, "__val__": b'"node121_val"'},
                    },
                    "node13": {"__acl__": {"aa": "cdrwa"}, "__val__": b"{}"},
                },
            ),
            ("/node1/node11", {"__acl__": {"aa": "cdrwa", "cc": "r"}, "__val__": b'"node11_val"'}),
            (
                "/node2",
                {
                    "__acl__": {"anyone": "cdrwa"},
                    "__val__": b'"node2_val"',
                    "node21": {"__acl__": {"anyone": "cdrwa"}, "__val__": b'"node21_val"'},
                    "node22": {"__acl__": {"aa": "rwa"}, "__val__": b"{}"},
                },
            ),
            (
                "/node3",
                {
                    "__acl__": {"aa": "rwca", "cc": "r"},
                    "__val__": b"{}",
                    "node31": {
                        "__acl__": {"aa": "rwca", "cc": "r"},
                        "__val__": b"{}",
                        "node311": {
                            "__acl__": {"aa": "rwca", "cc": "r"},
                            "__val__": b"{}",
                            "node3111": {"__acl__": {"aa": "rwca", "cc": "r"}, "__val__": b"{}"},
                            "node3112": {"__acl__": {"aa": "rwca", "cc": "r"}, "__val__": b"{}"},
                        },
                    },
                },
            ),
        )

        for path, expected_rst in valid_cases:
            rst = k3zkutil.export_hierarchy(zkcli, path)
            # Remove zookeeper system node for version-agnostic comparison
            # (ZK 3.9 adds different system nodes than older versions)
            if "zookeeper" in rst:
                del rst["zookeeper"]
            if "zookeeper" in expected_rst:
                expected_rst = dict(expected_rst)  # copy to avoid modifying original
                del expected_rst["zookeeper"]
            self.assertEqual(rst, expected_rst)

        zkcli.stop()


class TestWait(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        k3utdocker.pull_image(zk_tag)

    def setUp(self):
        k3utdocker.create_network()
        k3utdocker.start_container(
            zk_name,
            zk_tag,
            port_bindings={2181: 21811},
        )

        self.zk = wait_for_zk("127.0.0.1:21811")

        dd("start zk-test in docker")

    def tearDown(self):
        self.zk.stop()
        k3utdocker.remove_container(zk_name)

    def test_wait_absent(self):
        for wait_time in (
            -1,
            0.0,
            0.1,
            1,
        ):
            dd("no node wait:", wait_time)

            with k3ut.Timer() as t:
                k3zkutil.wait_absent(self.zk, "a", wait_time)
                self.assertAlmostEqual(0, t.spent(), delta=0.2)

    def test_wait_absent_no_timeout(self):
        def _del():
            time.sleep(1)
            self.zk.delete("a")

        for kwargs in (
            {},
            {"timeout": None},
        ):
            self.zk.create("a")
            th = k3thread.daemon(target=_del)

            with k3ut.Timer() as t:
                k3zkutil.wait_absent(self.zk, "a", **kwargs)
                self.assertAlmostEqual(1, t.spent(), delta=0.1)

            th.join()

    def test_wait_absent_timeout(self):
        self.zk.create("a")

        for wait_time in (
            -1,
            0.0,
            0.1,
            1,
        ):
            dd("node present wait:", wait_time)
            expected = max([0, wait_time])

            with k3ut.Timer() as t:
                self.assertRaises(k3zkutil.ZKWaitTimeout, k3zkutil.wait_absent, self.zk, "a", timeout=wait_time)
                self.assertAlmostEqual(expected, t.spent(), delta=0.2)

        self.zk.delete("a")

    def test_wait_absent_delete_node(self):
        delete_after = 0.2

        for wait_time in (
            0.5,
            1,
        ):
            dd("node present wait:", wait_time)

            self.zk.create("a")

            def _del():
                time.sleep(delete_after)
                self.zk.delete("a")

            th = k3thread.daemon(target=_del)
            with k3ut.Timer() as t:
                k3zkutil.wait_absent(self.zk, "a", wait_time)
                self.assertAlmostEqual(delete_after, t.spent(), delta=0.1)

            th.join()

    def test_wait_absent_change_node(self):
        self.zk.create("a")

        change_after = 0.2

        for wait_time in (
            0.5,
            1,
        ):
            dd("node present wait:", wait_time)
            expected = max([0, wait_time])

            def _change():
                time.sleep(change_after)
                self.zk.set("a", b"bbb")

            th = k3thread.daemon(target=_change)
            with k3ut.Timer() as t:
                self.assertRaises(k3zkutil.ZKWaitTimeout, k3zkutil.wait_absent, self.zk, "a", timeout=wait_time)
                self.assertAlmostEqual(expected, t.spent(), delta=0.1)

            th.join()

        self.zk.delete("a")

    def test_wait_absent_connection_lost(self):
        self.zk.create("a")

        def _close():
            time.sleep(0.3)
            self.zk.stop()

        th = k3thread.daemon(target=_close)

        with k3ut.Timer() as t:
            self.assertRaises(ConnectionClosedError, k3zkutil.wait_absent, self.zk, "a")
            self.assertAlmostEqual(0.3, t.spent(), delta=0.1)

        th.join()

    def test_get_next_no_version(self):
        cases = (
            -1,
            0.0,
            0.1,
            1,
        )

        for timeout in cases:
            self.zk.create("a", b"a-val")

            with k3ut.Timer() as t:
                k3zkutil.get_next(self.zk, "a", timeout=timeout, version=-1)
                self.assertAlmostEqual(0, t.spent(), delta=0.2)

            with k3ut.Timer() as t:
                k3zkutil.get_next(self.zk, "a", timeout=timeout)
                self.assertAlmostEqual(0, t.spent(), delta=0.2)

            self.zk.delete("a")

    def test_get_next_timeout(self):
        cases = (
            -1,
            0.0,
            0.2,
            1,
        )

        for timeout in cases:
            expected = max([timeout, 0])
            self.zk.create("a", b"a-val")

            with k3ut.Timer() as t:
                self.assertRaises(k3zkutil.ZKWaitTimeout, k3zkutil.get_next, self.zk, "a", timeout=timeout, version=0)
                self.assertAlmostEqual(expected, t.spent(), delta=0.2)

            self.zk.delete("a")

    def test_get_next_changed(self):
        cases = (
            0.4,
            1,
        )

        def _set_a():
            self.zk.set("a", b"changed")

        for timeout in cases:
            self.zk.create("a", b"a-val")
            th = k3thread.daemon(target=_set_a, after=0.3)

            with k3ut.Timer() as t:
                val, zstat = k3zkutil.get_next(self.zk, "a", timeout=timeout, version=0)
                self.assertAlmostEqual(0.3, t.spent(), delta=0.2)
                self.assertEqual(b"changed", val)
                self.assertEqual(1, zstat.version)

            th.join()
            self.zk.delete("a")

    def test_get_next_changed_but_unsatisfied(self):
        cases = (
            0.4,
            1,
        )

        def _set_a():
            self.zk.set("a", b"changed")

        for timeout in cases:
            self.zk.create("a", b"a-val")
            th = k3thread.daemon(target=_set_a, after=0.3)

            with k3ut.Timer() as t:
                self.assertRaises(k3zkutil.ZKWaitTimeout, k3zkutil.get_next, self.zk, "a", timeout=timeout, version=5)
                self.assertAlmostEqual(timeout, t.spent(), delta=0.2)

            th.join()
            self.zk.delete("a")

    def test_get_next_deleted(self):
        cases = (
            0.4,
            1,
        )

        def _del_a():
            self.zk.delete("a")

        for timeout in cases:
            self.zk.create("a", b"a-val")
            th = k3thread.daemon(target=_del_a, after=0.3)

            with k3ut.Timer() as t:
                self.assertRaises(NoNodeError, k3zkutil.get_next, self.zk, "a", timeout=timeout, version=0)
                self.assertAlmostEqual(0.3, t.spent(), delta=0.2)

            th.join()

    def test_get_next_conn_lost(self):
        self.zk.create("a", b"a-val")
        th = k3thread.daemon(target=self.zk.stop, after=0.3)

        with k3ut.Timer() as t:
            self.assertRaises(ConnectionClosedError, k3zkutil.get_next, self.zk, "a", timeout=1, version=0)
            self.assertAlmostEqual(0.3, t.spent(), delta=0.2)

        th.join()



import unittest

from k3confloader import conf
import k3ut
import k3zkutil

dd = k3ut.dd


class TestZKConf(unittest.TestCase):
    def test_specified(self):
        c = k3zkutil.ZKConf(
            hosts="hosts",
            tx_dir="tx_dir/",
            record_dir="record_dir/",
            seq_dir="seq_dir/",
            lock_dir="lock_dir/",
            node_id="node_id",
            auth=("digest", "a", "b"),
            acl=(("foo", "bar", "cd"), ("xp", "123", "cdrwa")),
        )

        self.assertEqual("hosts", c.hosts())
        self.assertEqual("tx_dir/", c.tx_dir())
        self.assertEqual("record_dir/", c.record_dir())
        self.assertEqual("seq_dir/", c.seq_dir())
        self.assertEqual("lock_dir/", c.lock_dir())
        self.assertEqual("node_id", c.node_id())
        self.assertEqual(("digest", "a", "b"), c.auth())

        self.assertEqual((("foo", "bar", "cd"), ("xp", "123", "cdrwa")), c.acl())

        self.assertEqual("lock_dir/", c.lock())
        self.assertEqual("lock_dir/a", c.lock("a"))
        self.assertEqual("record_dir/", c.record())
        self.assertEqual("record_dir/a", c.record("a"))
        self.assertEqual("tx_dir/alive/", c.tx_alive())
        self.assertEqual("tx_dir/alive/a", c.tx_alive("a"))
        self.assertEqual("tx_dir/alive/0000000001", c.tx_alive(1))
        self.assertEqual("tx_dir/state/", c.tx_state())
        self.assertEqual("tx_dir/state/a", c.tx_state("a"))
        self.assertEqual("tx_dir/state/0000000001", c.tx_state(1))
        self.assertEqual("tx_dir/journal/", c.journal())
        self.assertEqual("tx_dir/journal/a", c.journal("a"))
        self.assertEqual("tx_dir/journal/journal_id0000000001", c.journal(1))
        self.assertEqual("tx_dir/journal_id_set", c.journal_id_set())
        self.assertEqual("tx_dir/txid_maker", c.txid_maker())
        self.assertEqual("seq_dir/", c.seq())
        self.assertEqual("seq_dir/a", c.seq("a"))

        self.assertEqual(
            k3zkutil.make_kazoo_digest_acl((("foo", "bar", "cd"), ("xp", "123", "cdrwa"))), c.kazoo_digest_acl()
        )
        self.assertEqual(("digest", "a:b"), c.kazoo_auth())

    def test_default(self):
        old = (
            conf.zk_hosts,
            conf.zk_tx_dir,
            conf.zk_record_dir,
            conf.zk_lock_dir,
            conf.zk_node_id,
            conf.zk_auth,
            conf.zk_acl,
        )

        conf.zk_hosts = "HOSTS"
        conf.zk_tx_dir = "TX_DIR/"
        conf.zk_record_dir = "RECORD_DIR/"
        conf.zk_seq_dir = "SEQ_DIR/"
        conf.zk_lock_dir = "LOCK_DIR/"
        conf.zk_node_id = "NODE_ID"
        conf.zk_auth = ("DIGEST", "A", "B")
        conf.zk_acl = (("FOO", "BAR", "CD"), ("XP", "123", "CDRWA"))

        c = k3zkutil.ZKConf()

        self.assertEqual("HOSTS", c.hosts())
        self.assertEqual("TX_DIR/", c.tx_dir())
        self.assertEqual("RECORD_DIR/", c.record_dir())
        self.assertEqual("SEQ_DIR/", c.seq_dir())
        self.assertEqual("LOCK_DIR/", c.lock_dir())
        self.assertEqual("NODE_ID", c.node_id())
        self.assertEqual(("DIGEST", "A", "B"), c.auth())

        self.assertEqual((("FOO", "BAR", "CD"), ("XP", "123", "CDRWA")), c.acl())

        self.assertEqual("LOCK_DIR/", c.lock())
        self.assertEqual("LOCK_DIR/a", c.lock("a"))
        self.assertEqual("RECORD_DIR/", c.record())
        self.assertEqual("RECORD_DIR/a", c.record("a"))
        self.assertEqual("SEQ_DIR/", c.seq())
        self.assertEqual("SEQ_DIR/a", c.seq("a"))
        self.assertEqual("TX_DIR/alive/", c.tx_alive())
        self.assertEqual("TX_DIR/alive/a", c.tx_alive("a"))
        self.assertEqual("TX_DIR/alive/0000000001", c.tx_alive(1))
        self.assertEqual("TX_DIR/state/", c.tx_state())
        self.assertEqual("TX_DIR/state/a", c.tx_state("a"))
        self.assertEqual("TX_DIR/state/0000000001", c.tx_state(1))
        self.assertEqual("TX_DIR/journal/", c.journal())
        self.assertEqual("TX_DIR/journal/a", c.journal("a"))
        self.assertEqual("TX_DIR/journal/journal_id0000000001", c.journal(1))
        self.assertEqual("TX_DIR/journal_id_set", c.journal_id_set())
        self.assertEqual("TX_DIR/txid_maker", c.txid_maker())

        self.assertEqual(
            k3zkutil.make_kazoo_digest_acl((("FOO", "BAR", "CD"), ("XP", "123", "CDRWA"))), c.kazoo_digest_acl()
        )
        self.assertEqual(("DIGEST", "A:B"), c.kazoo_auth())

        (
            conf.zk_hosts,
            conf.zk_tx_dir,
            conf.zk_record_dir,
            conf.zk_lock_dir,
            conf.zk_node_id,
            conf.zk_auth,
            conf.zk_acl,
        ) = old



import logging

from kazoo.exceptions import BadVersionError

import k3txutil

from .zkconf import kazoo_client_ext
from . import zkutil

logger = logging.getLogger(__name__)


def cas_loop(zkclient, path, json=True):
    # while True:
    #     curr_val, zstat = zkclient.get(path)
    # new_val = curr_val + ':foo'
    # try:
    #     zkclient.set(path, new_val, version=zstat.version)
    # except BadVersionError as e:
    #     continue
    # else:
    #     return
    """
    A helper generator for doing CAS(check and set or compare and swap) on zk.
    See [CAS](https://en.wikipedia.org/wiki/Compare-and-swap)

    A general CAS loop is like following(check the version when update):
    :param zkclient: is a `KazooClient` instance connected to zk.
    It can also be a string, in which case it is treated as a comma separated
    hosts list, and a `zkclient` is created with default setting.
    It can also be a `dict` or an instance of `ZKConf`, in which case it create
    a `zkclient` with `ZKConf` defined setting.
    :param path: is the zk-node path to get and set.
    :param json: whether to do a json load after reading the value from zk and to do a json dump
    before updating the value to zk.
    :return: a generator yields a `tuple` of 2 element:
    """
    zkclient, owning_zk = kazoo_client_ext(zkclient, json=json)

    def setter(path, val, zstat):
        zkclient.set(path, val, version=zstat.version)

    try:
        for curr in k3txutil.cas_loop(zkclient.get, setter, args=(path,), conflicterror=BadVersionError):
            yield curr
    finally:
        if owning_zk:
            zkutil.close_zk(zkclient)
